# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Documentation control decorators."""
from typing import Iterable, Optional, TypeVar

T = TypeVar("T")

_DEPRECATED = "_tf_docs_deprecated"


def set_deprecated(obj: T) -> T:
  """Explicitly tag an object as deprecated for the doc generator."""
  setattr(obj, _DEPRECATED, None)
  return obj


def is_deprecated(obj) -> bool:
  return hasattr(obj, _DEPRECATED)


_INHERITABLE_HEADER = "_tf_docs_inheritable_header"


def inheritable_header(text: str):

  def _wrapped(obj):
    setattr(obj, _INHERITABLE_HEADER, text)
    return obj

  return _wrapped


def get_inheritable_header(obj) -> Optional[str]:
  return getattr(obj, _INHERITABLE_HEADER, None)


header = inheritable_header
get_header = get_inheritable_header


_NO_SEARCH_HINTS = "_tf_docs_no_search_hints"



def hide_from_search(obj: T) -> T:
  """Marks an object so metadata search hints will not be included on it's page.

  The page is set to "noindex" to hide it from search.

  Note: This only makes sense to apply to functions, classes and modules.
  Constants, and methods do not get their own pages.

  Args:
    obj: the object to hide.

  Returns:
    The object.
  """
  setattr(obj, _NO_SEARCH_HINTS, None)
  return obj


def should_hide_from_search(obj) -> bool:
  """Returns true if metadata search hints should not be included."""
  return hasattr(obj, _NO_SEARCH_HINTS)


_CUSTOM_PAGE_BUILDER_CLS = "_tf_docs_custom_page_builder_cls"


def set_custom_page_builder_cls(obj, cls):
  """Replace most of the generated page with custom content."""
  setattr(obj, _CUSTOM_PAGE_BUILDER_CLS, cls)


def get_custom_page_builder_cls(obj):
  """Gets custom page content if available."""
  return getattr(obj, _CUSTOM_PAGE_BUILDER_CLS, None)


_DO_NOT_DOC = "_tf_docs_do_not_document"


def do_not_generate_docs(obj: T) -> T:
  """A decorator: Do not generate docs for this object.

  For example the following classes:

  ```
  class Parent(object):
    def method1(self):
      pass
    def method2(self):
      pass

  class Child(Parent):
    def method1(self):
      pass
    def method2(self):
      pass
  ```

  Produce the following api_docs:

  ```
  /Parent.md
    # method1
    # method2
  /Child.md
    # method1
    # method2
  ```

  This decorator allows you to skip classes or methods:

  ```
  @do_not_generate_docs
  class Parent(object):
    def method1(self):
      pass
    def method2(self):
      pass

  class Child(Parent):
    @do_not_generate_docs
    def method1(self):
      pass
    def method2(self):
      pass
  ```

  This will only produce the following docs:

  ```
  /Child.md
    # method2
  ```

  Note: This is implemented by adding a hidden attribute on the object, so it
  cannot be used on objects which do not allow new attributes to be added. So
  this decorator must go *below* `@property`, `@classmethod`,
  or `@staticmethod`:

  ```
  class Example(object):
    @property
    @do_not_generate_docs
    def x(self):
      return self._x
  ```

  Args:
    obj: The object to hide from the generated docs.

  Returns:
    obj
  """
  setattr(obj, _DO_NOT_DOC, None)
  return obj


_DO_NOT_DOC_INHERITABLE = "_tf_docs_do_not_doc_inheritable"


def do_not_doc_inheritable(obj: T) -> T:
  """A decorator: Do not generate docs for this method.

  This version of the decorator is "inherited" by subclasses. No docs will be
  generated for the decorated method in any subclass. Even if the sub-class
  overrides the method.

  For example, to ensure that `method1` is **never documented** use this
  decorator on the base-class:

  ```
  class Parent(object):
    @do_not_doc_inheritable
    def method1(self):
      pass
    def method2(self):
      pass

  class Child(Parent):
    def method1(self):
      pass
    def method2(self):
      pass
  ```
  This will produce the following docs:

  ```
  /Parent.md
    # method2
  /Child.md
    # method2
  ```

  When generating docs for a class's arributes, the `__mro__` is searched and
  the attribute will be skipped if this decorator is detected on the attribute
  on any class in the `__mro__`.

  Note: This is implemented by adding a hidden attribute on the object, so it
  cannot be used on objects which do not allow new attributes to be added. So
  this decorator must go *below* `@property`, `@classmethod`,
  or `@staticmethod`:

  ```
  class Example(object):
    @property
    @do_not_doc_inheritable
    def x(self):
      return self._x
  ```

  Args:
    obj: The class-attribute to hide from the generated docs.

  Returns:
    obj
  """
  setattr(obj, _DO_NOT_DOC_INHERITABLE, None)
  return obj


_FOR_SUBCLASS_IMPLEMENTERS = "_tf_docs_tools_for_subclass_implementers"


def for_subclass_implementers(obj: T) -> T:
  """A decorator: Only generate docs for this method in the defining class.

  Also group this method's docs with and `@abstractmethod` in the class's docs.

  No docs will generated for this class attribute in sub-classes.

  The canonical use case for this is `tf.keras.layers.Layer.call`: It's a
  public method, essential for anyone implementing a subclass, but it should
  never be called directly.

  Works on method, or other class-attributes.

  When generating docs for a class's arributes, the `__mro__` is searched and
  the attribute will be skipped if this decorator is detected on the attribute
  on any **parent** class in the `__mro__`.

  For example:

  ```
  class Parent(object):
    @for_subclass_implementers
    def method1(self):
      pass
    def method2(self):
      pass

  class Child1(Parent):
    def method1(self):
      pass
    def method2(self):
      pass

  class Child2(Parent):
    def method1(self):
      pass
    def method2(self):
      pass
  ```

  This will produce the following docs:

  ```
  /Parent.md
    # method1
    # method2
  /Child1.md
    # method2
  /Child2.md
    # method2
  ```

  Note: This is implemented by adding a hidden attribute on the object, so it
  cannot be used on objects which do not allow new attributes to be added. So
  this decorator must go *below* `@property`, `@classmethod`,
  or `@staticmethod`:

  ```
  class Example(object):
    @property
    @for_subclass_implementers
    def x(self):
      return self._x
  ```

  Args:
    obj: The class-attribute to hide from the generated docs.

  Returns:
    obj
  """
  setattr(obj, _FOR_SUBCLASS_IMPLEMENTERS, None)
  return obj


do_not_doc_in_subclasses = for_subclass_implementers

_DOC_PRIVATE = "_tf_docs_doc_private"


def doc_private(obj: T) -> T:
  """A decorator: Generates docs for private methods/functions.

  For example:

  ```
  class Try:

    @doc_controls.doc_private
    def _private(self):
      ...
  ```

  As a rule of thumb, private(beginning with `_`) methods/functions are
  not documented.

  This decorator allows to force document a private method/function.

  Args:
    obj: The class-attribute to hide from the generated docs.

  Returns:
    obj
  """

  setattr(obj, _DOC_PRIVATE, None)
  return obj


def should_doc_private(obj) -> bool:
  return hasattr(obj, _DOC_PRIVATE)


_DOC_IN_CURRENT_AND_SUBCLASSES = "_tf_docs_doc_in_current_and_subclasses"


def doc_in_current_and_subclasses(obj: T) -> T:
  """Overrides `do_not_doc_in_subclasses` decorator.

  If this decorator is set on a child class's method whose parent's method
  contains `do_not_doc_in_subclasses`, then that will be overridden and the
  child method will get documented. All classes inheriting from the child will
  also document that method.

  For example:

  ```
  class Parent:
    @do_not_doc_in_subclasses
    def method1(self):
      pass
    def method2(self):
      pass

  class Child1(Parent):
    @doc_in_current_and_subclasses
    def method1(self):
      pass
    def method2(self):
      pass

  class Child2(Parent):
    def method1(self):
      pass
    def method2(self):
      pass

  class Child11(Child1):
    pass
  ```

  This will produce the following docs:

  ```
  /Parent.md
    # method1
    # method2
  /Child1.md
    # method1
    # method2
  /Child2.md
    # method2
  /Child11.md
    # method1
    # method2
  ```

  Args:
    obj: The class-attribute to hide from the generated docs.

  Returns:
    obj
  """

  setattr(obj, _DOC_IN_CURRENT_AND_SUBCLASSES, None)
  return obj


def should_skip(obj) -> bool:
  """Returns true if docs generation should be skipped for this object.

  Checks for the `do_not_generate_docs` or `do_not_doc_inheritable` decorators.

  Args:
    obj: The object to document, or skip.

  Returns:
    True if the object should be skipped
  """
  if isinstance(obj, type):
    # For classes, only skip if the attribute is set on _this_ class.
    if _DO_NOT_DOC in obj.__dict__:
      return True
    else:
      return False

  # Unwrap fget if the object is a property
  if isinstance(obj, property):
    obj = obj.fget

  return hasattr(obj, _DO_NOT_DOC) or hasattr(obj, _DO_NOT_DOC_INHERITABLE)


def _unwrap_func(obj):
  # Unwrap fget if the object is a property or static method or classmethod.
  if isinstance(obj, property):
    return obj.fget

  if isinstance(obj, (classmethod, staticmethod)):
    return obj.__func__

  return obj


def _cls_attr_has_tag(cls, attr, tag):
  """Check if a class attribute `attr` is decorated with `dec`."""
  # Use __dict__, it doesn't go up the __mro__ like getattr.
  obj = cls.__dict__.get(attr, None)
  if obj is None:
    return False
  obj = _unwrap_func(obj)

  if isinstance(obj, type):
    # The attribute is a class. Check __dict__ to see if the attribute is set
    # on _this_ class, not its parents.
    if tag in obj.__dict__:
      return True
    else:
      return False

  return hasattr(obj, tag)


def should_skip_class_attr(cls, name):
  """Returns true if docs should be skipped for this class attribute.

  Args:
    cls: The class the attribute belongs to.
    name: The name of the attribute.

  Returns:
    True if the attribute should be skipped.
  """
  # Get the object with standard lookup, from the nearest
  # defining parent.
  try:
    obj = getattr(cls, name)
  except AttributeError:
    # This can fail for a variety of reasons. Always skip if `getattr` fails.
    return True

  # Unwrap fget if the object is a property
  obj = _unwrap_func(obj)

  # Skip if the object is decorated with `do_not_generate_docs` or
  # `do_not_doc_inheritable`
  if should_skip(obj):
    return True

  classes = getattr(cls, "__mro__", [cls])

  # Find where all the decorators turn docs on and off.
  # All these lists contain `(level, skip)` pairs.
  for_subclass_levels = [
      # The [1:] is because `for_subclass_implementers` turns off docs
      # one level down (and you don't want to consider level -1).
      (i, True)
      for (i, mro_cls) in enumerate(classes[1:])
      if _cls_attr_has_tag(mro_cls, name, _FOR_SUBCLASS_IMPLEMENTERS)
  ]
  not_below_levels = [
      (i, True)
      for (i, mro_cls) in enumerate(classes)
      if _cls_attr_has_tag(mro_cls, name, _DO_NOT_DOC_INHERITABLE)
  ]
  doc_below_levels = [
      (i, False)
      for (i, mro_cls) in enumerate(classes)
      if _cls_attr_has_tag(mro_cls, name, _DOC_IN_CURRENT_AND_SUBCLASSES)
  ]

  all_levels = not_below_levels + for_subclass_levels + doc_below_levels
  if all_levels:
    # Find the lowest `(level, skip)` pair, and return `skip`
    return min(all_levels)[1]

  # No decorators --> don't skip
  return False


def decorate_all_class_attributes(decorator, cls, skip: Iterable[str]):
  """Applies `decorator` to every attribute defined in `cls`.

  Args:
    decorator: The decorator to apply.
    cls: The class to apply the decorator to.
    skip: A collection of attribute names that the decorator should not be
      aplied to.
  """
  skip = frozenset(skip)
  class_contents = list(cls.__dict__.items())

  for name, obj in class_contents:
    if name in skip:
      continue

    # Otherwise, exclude from documentation.
    if isinstance(obj, property):
      obj = obj.fget

    if isinstance(obj, (staticmethod, classmethod)):
      obj = obj.__func__

    try:
      decorator(obj)
    except AttributeError:
      pass
